package edu.columbia.c2b2.aracne;

import edu.columbia.c2b2.aracne.Util.Gene;
import edu.columbia.c2b2.aracne.Util.Sort_Gene;
import wb.data.Marker;
import wb.data.MicroarraySet;
import wb.data.MarkerSet;

import java.util.Map;
import java.util.Vector;
import java.util.Collections;
import java.util.Hashtable;

/**
 * Marker statistics used in ARACNE
 * User: mhall
 * User: manjunath at c2b2 dot columbia dot edu
 * Date: Aug 22, 2006
 * Time: 11:40:52 AM
 */
public class MarkerStats {

    Hashtable<Marker, MarkerStat> markerStats = new Hashtable<Marker, MarkerStat>();

    void computeMarkerBandwidth(MicroarraySet set) {
        int n = set.size();
        Vector<Double> data = new Vector<Double>(n);
        for (int i = 0; i < set.getMarkers().size(); i++) {

            double prop = 1.06; // Gaussian
            int dim = 1; // Dimension of data

            MarkerStat stat = markerStats.get(set.getMarkers().getMarker(i));
            double std_dev = Math.sqrt(stat.getVariance());
            data.clear();
            for (int j = 0; j < n; j++) {
                data.add((double) set.getValue(j, i));
            }

            Collections.sort(data);

            double iqr = Util.interQuartileRange(data, n);
            double iqrSig = .7413 * iqr; // find interquartile range sigma est.

            if (iqrSig == 0) {
                iqrSig = std_dev;
            }
            double sig = Math.min(std_dev, iqrSig);

            // Computing bandwidth
            double h = prop * sig * Math.pow((double) n, ((double) -1) / ((double) (4 + dim)));
            stat.setBandwidth(h);
        }
    }

    void computeMarkerRanks(MicroarraySet set) {
        int n = set.size();
        for (int i = 0; i < set.getMarkers().size(); i++) {

            MarkerStat stat = markerStats.get(set.getMarkers().getMarker(i));

            Vector<Gene> data = new Vector<Gene>(n);
            for (int j = 0; j < n; j++) {
                Gene g = new Gene(j, (double) set.getValue(j, i));
                data.add(g);
            }
            Collections.sort(data, new Sort_Gene());
            for (int j = 0; j < n; j++)
                data.get(j).xi = j;
            stat.setRankedValues(data);
        }
    }
    
    void computeMarkerVariance(MicroarraySet set) {
        MarkerSet markers = set.getMarkers();
        for (int i = 0; i < markers.size(); i++) {
            double var = variance(i, set);
            Marker marker = markers.getMarker(i);
            MarkerStats.MarkerStat markerStat = markerStats.get(marker);
            if (markerStat == null) {
                markerStats.put(marker, new MarkerStat(true, var));
            } else {
                markerStat.setVariance(var);
            }
        }
    }

    double variance(int m, MicroarraySet set) {

        double var;
        int n = set.size();
        double s = 0;
        double ss = 0;
        for (int i = 0; i < n; i++) {
            double v = set.getValue(i, m);
            s += v;
            ss += v * v;
        }
        var = (ss - s * s / n) / (n - 1);
        return var;
    }

    MarkerStat getMarkerStat(Marker marker) {
        MarkerStats.MarkerStat markerStat = markerStats.get(marker);
        if (markerStat != null) {
            return markerStat;
        } else {
            markerStat = new MarkerStat();
            markerStats.put(marker, markerStat);
            return markerStat;
        }
    }

    public int Get_Num_Active_Markers() {
        int count = 0;
        for (Map.Entry<Marker, MarkerStat> entry : markerStats.entrySet()) {
            if (!entry.getValue().isEnabled()) {
                count++;
            }
        }
        return count;
    }

    public static class MarkerStat {

        double variance = 0.0;
        double bandwidth = 0.0;
        boolean enabled = true;
        boolean control = false;
        Vector<Gene> rankedValues;
        
        public MarkerStat() {
        }

        public MarkerStat(boolean control, boolean enabled, double variance) {
            this.control = control;
            this.enabled = enabled;
            this.variance = variance;
        }

        public MarkerStat(boolean enabled, double variance) {
            this.enabled = enabled;
            this.variance = variance;
        }

        public boolean isEnabled() {
            return enabled;
        }

        public void setEnabled(boolean enabled) {
            this.enabled = enabled;
        }

        public double getVariance() {
            return variance;
        }

        public void setVariance(double variance) {
            this.variance = variance;
        }

        public void setBandwidth(double bandwidth) {
            this.bandwidth = bandwidth;
        }

        public double getBandwidth() {
            return bandwidth;
        }

        public void setRankedValues(Vector<Gene> r) {
            rankedValues = r;
        }
        
        public Vector<Gene> getRankedValues() {
            return rankedValues;
        }
        
        public boolean isControl() {
            return control;
        }

        public void setControl(boolean control) {
            this.control = control;
        }
    }
}